#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
:class:`metabeyond.hints.Hint` documentation decorators for other components in this framework.
"""
__all__ = [
    "slotted",
    "discovery",
    "class_only_decorator",
    "function_only_decorator",
    "method_only_decorator",
    "decorator",
    "interface",
    "abstract_class",
]

from metabeyond import hints

# pylint: disable=C0103
#: Decorates any class that uses `__slots__` internally.
#:
#: See https://docs.python.org/3/reference/datamodel.html#object.__slots__
slotted = hints.Hint(
    "This class is slotted, so does not allow dynamic attribute injection without subclassing."
)

# pylint: disable=C0103
#: Decorates any element that is used when discovery at runtime.
discovery = hints.Hint(
    "This element has the ability to be discovered when performing introspection at runtime."
)

# pylint: disable=C0103
#: Marks an element as a decorator that can only decorate classes.
class_only_decorator = hints.Hint("This can only decorate classes.")

# pylint: disable=C0103
#: Marks an element as a decorator that can only decorate functions.
function_only_decorator = hints.Hint("This can only decorate functions.")

# pylint: disable=C0103
#: Marks an element as a decorator that can only decorate methods (functions within a class context).
method_only_decorator = hints.Hint("This can only decorate methods.")

# pylint: disable=C0103
#: Mark an element as being a decorator for anything.
decorator = hints.decorator

# pylint: disable=C0103
#: Mark an element as being an interface.
interface = hints.interface

# pylint: disable=C0103
#: Marks an abstract class.
abstract_class = hints.Hint('This is an abstract class.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Dependency injector base class.
"""
__all__ = ['BaseDependencyInjector']

import abc
from typing import Any, Iterator

from ricochet import dependency, exceptions
from ricochet.utils import hints


@hints.interface
class BaseDependencyInjector(abc.ABC):
    """
    Interface specification for an entity with the ability to resolve and inject dependencies into objects.
    """

    @abc.abstractmethod
    def resolve_dependencies(self, object: Any) -> None:
        """
        Resolves any autowired dependencies in the given object.

        Args:
            object:
                The object to resolve dependencies for.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_dependencies(self, object: Any) -> Iterator[dependency.BaseDependencyDescriptor]:
        """
        Returns an iterator across any dependencies in the given object that need to be resolved.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def wire_in_objectlet(self, object: Any, name: str, objectlet: Any) -> None:
        """
        Wire in the given objectlet to the given object under the given name.

        Args:
            object:
                the object to inject the dependency into.
            name:
                the name of the objectlet dependency.
            objectlet:
                the objectlet dependency itself.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_objectlet_from_dependency(
        self, dependency_descriptor: dependency.ObjectletDependencyDescriptor
    ) -> Any:
        """
        Takes an objectlet dependency descriptor and attempts to get the objectlet from a registry matching that
        definition.

        Args:
            dependency_descriptor:
                The description of the dependency to locate.

        Returns:
            The dependency.
        """
        raise exceptions.AbstractError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Wraps a prototype and singleton registry and provides a managed unified interface to both at the same time.
"""
__all__ = ["BaseDualRegistryMediator"]

import abc
from typing import Iterator, Any, Type

from ricochet import exceptions, scopes
from ricochet.utils import hints
from . import base_prototype_registry, base_singleton_registry


@hints.abstract_class
class BaseDualRegistryMediator(
    base_prototype_registry.BasePrototypeRegistry,
    base_singleton_registry.BaseSingletonRegistry,
    abc.ABC,
):
    """
    This abstracts prototype and singleton objectlet_registries to provide a system that allows access to both types of objectlet_registries
    using the same methods in an expected way. This functions by overriding conflicting implementations.
    """

    def __init__(self):
        super().__init__()

    def contains_objectlet_named(self, canonical_name: str) -> bool:
        """
        Args:
            canonical_name:
                The canonical object name. This is the primary name the object is registered under and resolved by.
                This implements no form of alias lookup.

        Returns:
            `True` if a prototype or singleton with the canonical name is registered, `False` if not.
        """
        singleton = self.contains_singleton_named(canonical_name)
        prototype = self.contains_prototype_named(canonical_name)

        return singleton or prototype

    def contains_objectlets_typed(self, type: Type[Any]) -> bool:
        """
        Args:
            type:
                A class that a prototype or singleton is expected to be derived from.

        Returns:
            `True` if any prototype or singleton derived from the given type exists, `False` if not.
        """
        singleton = self.contains_singletons_typed(type)
        prototype = self.contains_prototypes_typed(type)

        return singleton or prototype

    def count_objectlets_named(self, canonical_name: str) -> int:
        """
        Args:
            canonical_name:
                The canonical object name. This is the primary name the object is registered under and resolved by.
                This implements no form of alias lookup.

        Returns:
            A count of object instances, if the prototype or singleton exists for that name.

        Raises:
            NameError if no prototype or singleton exists with that name.
        """
        # We can safely make the assumption that
        # >>> assert singleton_count ^ prototype_count
        # if we use the basic implementation rules, but user implementations may do voodoo type shit, so
        # we may as well cater for that and it makes the operation terse.
        prototypes = self.count_instances_of_prototype_named(canonical_name)
        singleton = 1 if self.contains_singleton_named(canonical_name) else 0
        return prototypes + singleton

    def is_objectlet(self, candidate_object: Any) -> bool:
        """
        Args:
            candidate_object: The potential singleton or prototype instance to look for.

        Returns:
            `True` if this object is a prototype instance or a singleton object in this registry, or `False` otherwise.
        """
        return self.is_prototype_instance(candidate_object) or self.is_singleton(candidate_object)

    def get_all_canonical_objectlet_names(self) -> Iterator[str]:
        """
        Yields all object names. These are expected to be unique as long as the internal representations that are
        protected have not been tampered with directly.
        """
        yield from self.get_all_singleton_names()
        yield from self.get_all_prototype_names()

    ########################
    # Overridden functions #
    ########################

    def register_new_prototype(self, canonical_name: str, prototype: Any) -> None:
        """
        Registers the given object as a prototype under the given name.

        Args:
           canonical_name:
               The canonical prototype's name. This is the primary name the object is registered under and resolved by.
           prototype:
               The object itself to register.

        Throws:
           ValueError:
               if the prototype is `None`.
           NameError:
               if the prototype is already registered as a singleton or prototype.
        """
        if canonical_name in self.get_all_singleton_names():
            raise exceptions.ObjectletIsAlreadyRegisteredError(
                canonical_name, scopes.Scope.SINGLETON
            )
        elif canonical_name in self.get_all_prototype_names():
            raise exceptions.ObjectletIsAlreadyRegisteredError(
                canonical_name, scopes.Scope.PROTOTYPE
            )
        else:
            return super().register_new_prototype(canonical_name, prototype)

    def register_new_singleton(self, canonical_name: str, singleton: Any) -> None:
        """
        Registers the given object as a singleton under the given name.

        Args:
           canonical_name:
               The canonical singleton's name. This is the primary name the object is registered under and resolved by.
           singleton:
               The object itself to register.

        Throws:
           :class:`ValueError`:
               if the singleton is `None`.
           :class:`NameError`:
               if the singleton is already registered as a singleton or prototype
        """
        if canonical_name in self.get_all_singleton_names():
            raise exceptions.ObjectletIsAlreadyRegisteredError(
                canonical_name, scopes.Scope.SINGLETON
            )
        elif canonical_name in self.get_all_prototype_names():
            raise exceptions.ObjectletIsAlreadyRegisteredError(
                canonical_name, scopes.Scope.PROTOTYPE
            )
        else:
            return super().register_new_singleton(canonical_name, singleton)

    def get_objectlet(self, canonical_name: str) -> Any:
        singleton = self.get_singleton_named(canonical_name)
        if singleton is not None:
            return singleton

        prototype = self.create_instance_of_prototype(canonical_name)
        if prototype is not None:
            return prototype

        raise exceptions.ObjectletNotFoundError(canonical_name)

    def register_new_objectlet(self, canonical_name: str, object: Any, scope: scopes.Scope):
        if scope is scopes.Scope.SINGLETON:
            self.register_new_singleton(canonical_name, object)
        elif scope is scopes.Scope.PROTOTYPE:
            self.register_new_prototype(canonical_name, object)
        else:
            raise NotImplementedError(
                f'Scope {scope.name} is not implemented for this type of registry'
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base exceptions used in this framework.
"""
from ricochet import scopes


class RicochetError(RuntimeError):
    """
    Base runtime error for Ricochet.
    """

    def __init__(self, message: str, *args, **kwargs):
        error_message = message.format(*args, **kwargs)
        self.error_message = error_message
        super().__init__()

    def __repr__(self):
        return f'<{type(self).__name__}: {self.error_message!r}>'

    def __str__(self):
        return self.error_message


class AbstractError(RicochetError):
    """
    Raised if an accessed element is abstract.
    """

    def __init__(self):
        super().__init__('This element is abstract and should be overridden to be used')


class ObjectletManagementError(RicochetError):
    """
    Base for any objectlets that fail to load.
    """


class ObjectletNotFoundError(ObjectletManagementError):
    """
    Raised if an Objectlet cannot be resolved because it does not exist.
    """

    def __init__(self, objectlet_name: str):
        super().__init__('Objectlet {} does not exist in the registry', objectlet_name)


class ObjectletIsAlreadyRegisteredError(ObjectletManagementError):
    """
    Raised if an Objectlet under the given signature already exists.
    """

    def __init__(self, objectlet_name: str, scope: scopes.Scope):
        super().__init__(
            'Objectlet {} already exists as a {} in this registry',
            objectlet_name,
            scope.name.lower(),
        )


class ObjectletAliasIsAlreadyTakenError(ObjectletManagementError):
    """
    Raised when registering an alias if the alias is already taken.
    """

    def __init__(self, desired_alias: str, canonical_name: str, taken_by_name: str):
        super().__init__(
            'Cannot register alias {} to {} as it is already taken by {}',
            desired_alias,
            canonical_name,
            taken_by_name,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Prototype registry implementation that uses normal references.
"""
__all__ = ["PrototypeRegistry", "PrototypeMediator"]

import copy
import weakref
from typing import Any, KeysView, Optional, Type

from ricochet.objectlet_registries import base_prototype_registry
from ricochet.utils import logger_mixin, hints


@hints.slotted
class PrototypeMediator:
    """
    Internal wrapper for a prototype which handles maintaining counts of prototypes and producing new prototypes
    via a deep-copy. Since this is slotted to reduce overhead, no ABC will be made for this interface.

    Contains a logger to use which is passed from the containing prototype registry implementation to provide
    appropriate logging under a meaningful logger name.
    """

    __slots__ = ("logger", "prototype_name", "prototype_obj", "instances")

    def __init__(self, logger, prototype_name, prototype_obj):
        #: The logger to log with.
        self.logger = logger
        #: The name of the prototype this adapter wraps.
        self.prototype_name = prototype_name
        #: The actual prototype object itself.
        self.prototype_obj = prototype_obj
        #: Weak references pointing to the instances of the :attr:`prototype_obj` prototype. This automatically
        #: destroys references to any prototype instance that has been garbage collected, so it maintains itself.
        self.instances = weakref.WeakSet()

    def __len__(self):
        return len(self.instances)

    def make_new_instance(self):
        inst = copy.deepcopy(self.prototype_obj)
        self.logger.info(
            "Created new instance of prototype %s (#%s) at #%s",
            self.prototype_name,
            id(self.prototype_obj),
            id(inst),
        )

        self.instances.add(inst)
        return inst


class PrototypeRegistry(base_prototype_registry.BasePrototypeRegistry, logger_mixin.LoggerMixin):
    """
    Basic implementation of a Prototype registry. This generates objectlets from a given prototype under a specific
    identifier, aiming to be the simplest implementation possible.

    Warning:
        This produces prototypes using deep-copy internally. This will fail to produce correct prototypes on objects
        such as strings and ints which are internally optimised as reused pooled objects in CPython.

        It will also not work on metaclasses (i.e. the type of a type).

        Weak references are used internally to maintain a total count of prototype objectlets created. This means that
        it will fail for internal types such as :class:`object` which are slotted and immutable to weakref entries
        by default.
    """

    def __init__(self) -> None:
        super().__init__()
        self._prototypes = {}

    def contains_prototype_named(self, canonical_name: str) -> bool:
        return canonical_name in self._prototypes

    def contains_prototypes_typed(self, type: Type[Any]) -> bool:
        return any(
            True
            for prototype in self._prototypes.values()
            if isinstance(prototype.prototype_obj, type)
        )

    def create_instance_of_prototype(self, canonical_name: str) -> Optional[Any]:
        try:
            return self._prototypes[canonical_name].make_new_instance()
        except KeyError:
            raise NameError(f"Prototype {canonical_name} does not exist")

    def register_new_prototype(self, canonical_name: str, prototype: Any) -> None:
        if prototype is None:
            raise ValueError("Cannot register a NoneType prototype")
        elif canonical_name in self._prototypes:
            raise NameError("Prototype is already registered")
        else:
            self.logger.debug(
                "Registering prototype %r as %s-typed %r",
                canonical_name,
                type(prototype).__name__,
                prototype,
            )

            self._prototypes[canonical_name] = PrototypeMediator(
                self.logger, canonical_name, prototype
            )

    def is_prototype_instance(self, candidate_prototype_instance: Any) -> bool:
        return any(
            candidate_prototype_instance in prototype_header.instances
            for prototype_header in self._prototypes.values()
        )

    def count_instances_of_prototype_named(self, canonical_name: str) -> int:
        if canonical_name in self._prototypes:
            return len(self._prototypes[canonical_name])
        else:
            raise NameError(f"No prototype named {canonical_name!r} exists")

    def get_all_prototype_names(self) -> KeysView[str]:
        return self._prototypes.keys()

    @property
    def prototype_count(self) -> int:
        return len(self._prototypes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import typing


class Autowired:
    """
    Marks a class as being autowired.

        >>> foo: Autowired[int]
        >>> bar: Autowired
    """

    @classmethod
    def __class_getitem__(cls, item):
        """
        If we are type checking, then use the type hint directly as the type, but if we are
        running then we need to retain the class as the annotation or it will fail to be detected.
        """
        return typing.TYPE_CHECKING and item or cls



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Defines a base registry interface for a singleton registry.

**Singleton objects**:
    Global objects created once and shared. Often these refer to classes that may only ever have one instance,
    whereas in this framework, it refers to an object with a specific name. Zero or more managed components
    may store a reference to zero or more singleton instances. If one component changes the state of a singleton,
    then this change seen by all components that store a reference to the same singleton.
"""

__all__ = ["BaseSingletonRegistry"]

import abc
from typing import Any, Iterator, ItemsView, KeysView, Optional, Type, ValuesView

from ricochet import exceptions, scopes
from ricochet.objectlet_registries import base_registry


class BaseSingletonRegistry(base_registry.BaseRegistry):
    """
    Defines an object which maintains some collection of singleton objectlets. These are registered
    and then kept for the lifetime of the object unless otherwise specified, and are stored as a mapping
    of a string name to an objectlet instance. Thus, only one objectlet may have the same name.

    .. image:: resources_static/singleton-cardinality.png

    Singletons are shared among multiple components across the application context. If one component alters the
    state of a singleton, it will alter the state for all references, as singletons will be initialised once and shared.

    Note:
        Singleton objectlets are defined as objects that exist once per registry, and are provided as references
        to anything that requires the dependency.

    .. image:: resources_static/singleton-distribution.png

    Takes a generic TypeVar :attr:`SingletonWrapperT` describing the base type to output when queried for a singleton
    instance. If no wrapper is provided (i.e. the singleton is just output "as is"), this can be omitted, in which case,
    it will default to `Any`.

    Note:
        No implementation is expected or should attempt to implement lazy initialisation support.
        This is expected to be managed by a higher level abstraction that uses this registry internally.
    """

    @abc.abstractmethod
    def contains_singleton_named(self, canonical_name: str) -> bool:
        """
        Args:
            canonical_name:
                The canonical objectlet name. This is the primary name the objectlet is registered under and resolved
                by. This implements no form of alias lookup.

        Returns:
            `True` if this registry contains the singleton objectlet with the given name, or `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def contains_singletons_typed(self, type: Type[Any]) -> bool:
        """
        Args:
            type:
                A class that a singleton is expected to be derived from.

        Returns:
            `True` if this registry contains a singleton objectlet derived from the given type, or `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_singleton_named(self, canonical_name: str) -> Optional[Any]:
        """
        Args:
            canonical_name:
                The canonical singleton name. This is the primary name the objectlet is registered under and resolved
                by. This implements no form of alias lookup.

        Returns:
            The singleton with the given name, or `None` if it doesn't exist...
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_singletons_typed(self, type: Type[Any]) -> Iterator[Any]:
        """
        Returns:
             An iterator across any singleton objectlets derived from the given type that exist in
             this registry.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def register_new_singleton(self, canonical_name: str, singleton: Any) -> None:
        """
        Registers the given object as a singleton objectlet under the given name.

        Args:
            canonical_name:
                The canonical singleton's name. This is the primary name the objectlet is registered under and
                resolved by.
            singleton:
                The object itself to register.

        Throws:
            :class:`ValueError`:
                if the singleton object is `None`.
            :class:`NameError`:
                if the singleton is already registered.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def is_singleton(self, candidate_singleton: Any) -> bool:
        """
        Returns:
             `True` if the given object is a registered singleton, `False` otherwise.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_all_singletons(self) -> ItemsView[str, Any]:
        """
        Gets an immutable mapping of singleton names to singleton objectlet instances in this registry.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_all_singleton_names(self) -> KeysView[str]:
        """
        Gets an immutable key view of singleton names in this registry.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_all_singleton_instances(self) -> ValuesView[Any]:
        """
        Gets an immutable value view of singleton instances in this registry.
        """
        raise exceptions.AbstractError

    @property
    @abc.abstractmethod
    def singleton_count(self) -> int:
        """
        The number of objectlets registered as singletons.
        """
        raise exceptions.AbstractError

    def get_objectlet(self, canonical_name: str) -> Any:
        return self.get_singleton_named(canonical_name)

    def count_objectlets_named(self, canonical_name: str):
        return 1 if canonical_name in self.get_all_singleton_names() else 0

    def contains_objectlet_named(self, canonical_name: str) -> bool:
        return self.contains_singleton_named(canonical_name)

    def contains_objectlets_typed(self, type: Type[Any]) -> bool:
        return self.contains_singletons_typed(type)

    def is_objectlet(self, candidate_object: Any) -> bool:
        return self.is_singleton(candidate_object)

    def get_all_canonical_objectlet_names(self) -> Iterator[str]:
        yield from self.get_all_singleton_names()

    def register_new_objectlet(self, canonical_name: str, object: Any, scope: scopes.Scope) -> None:
        if scope is not scopes.Scope.SINGLETON:
            raise NotImplementedError('This class can only register singleton objects')
        else:
            self.register_new_singleton(canonical_name, object)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Objectlet scope.
"""
import enum


class Scope(enum.Enum):
    """Describes the scope of an objectlet."""

    #: Singleton object
    SINGLETON = enum.auto()
    #: Base object that would be used to generate prototypes.
    PROTOTYPE = enum.auto()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Defines base registry interfaces for Singleton and Prototype objectlet_registries. These are the core starting points for the
entire organisation implemented by this framework.

Keeping definitions simple, assuming you understand object orientation concepts:
    **Singletons**:
        Global objects created once and shared. Often these refer to classes that may only ever have one instance,
        whereas in this framework, it refers to an object with a specific name. Zero or more managed components
        may store a reference to zero or more singleton instances. If one component changes the state of a singleton,
        then this change seen by all components that store a reference to the same singleton.

    **Prototypes**:
        An object which is stored internally by the registry once. Each component that requests a reference to the
        prototype will receive some object created from the state of the prototype. Changing this object will not
        mutate the original prototype in any way, and thus provides a copy to every component that requires one.

        This is very similar to a normal instance, but instead of using a class to create an object, we use an
        existing object to derive a copy from.

        Likewise, these are much more difficult to track, as there is no implementation for tracking instances
        created from prototypes.

Registries manage a collection of objects. The objectlet_registries here contain implementations to replicate and distribute
singleton or prototype objects. These are used elsewhere in this framework for autowiring, etc.
"""
from .base_dual_registry_mediator import *
from .base_prototype_registry import *
from .base_registry import *
from .base_singleton_registry import *
from .impl.dual_registry_mediator import *
from .impl.prototype_registry import *
from .impl.singleton_registry import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Singleton registry implementation that uses normal references.
"""
__all__ = ["SingletonRegistry"]

from typing import Any, Dict, ItemsView, Iterator, KeysView, Optional, Type, ValuesView

from ricochet.objectlet_registries import base_singleton_registry
from ricochet.utils import logger_mixin


class SingletonRegistry(base_singleton_registry.BaseSingletonRegistry, logger_mixin.LoggerMixin):
    """
    Basic implementation of a Singleton registry. This provides access to any singleton objectlets as normal references,
    as the simplest implementation.
    """

    def __init__(self):
        super().__init__()
        self._singletons: Dict[str, Any] = {}

    def contains_singleton_named(self, canonical_name: str) -> bool:
        return canonical_name in self._singletons

    def contains_singletons_typed(self, type: Type[Any]) -> bool:
        iterator = self.get_singletons_typed(type)
        try:
            next(iterator)
        except StopIteration:
            return False
        else:
            return True

    def get_singleton_named(self, canonical_name: str) -> Optional[Any]:
        if canonical_name in self._singletons:
            return self._singletons[canonical_name]
        else:
            return None

    def get_singletons_typed(self, type: Type[Any]) -> Iterator[Any]:
        yield from (obj for obj in self._singletons.values() if isinstance(obj, type))

    def register_new_singleton(self, canonical_name: str, singleton: Any) -> None:
        if singleton is None:
            raise ValueError("Cannot register a NoneType singleton")
        elif canonical_name in self._singletons:
            raise NameError("Singleton is already registered")
        else:
            self.logger.debug(
                "Registering singleton %r from %s-typed %r",
                canonical_name,
                type(singleton).__name__,
                singleton,
            )
            self._singletons[canonical_name] = singleton

    def is_singleton(self, candidate_singleton: Any) -> bool:
        # This is expected to be a proxy only.
        return candidate_singleton in self._singletons.values()

    def get_all_singletons(self) -> ItemsView[str, Any]:
        return self._singletons.items()

    def get_all_singleton_names(self) -> KeysView[str]:
        return self._singletons.keys()

    def get_all_singleton_instances(self) -> ValuesView[Any]:
        return self._singletons.values()

    @property
    def singleton_count(self) -> int:
        return len(self._singletons)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Remark annotations.
"""
import inspect

from metabeyond import remarks

from ricochet import scopes


class Objectlet(remarks.Remark, constraint=inspect.isfunction):
    """
    Marks a method in a configuration that should be invoked to obtain an objeclet.

    Keyword Args:
        scope:
            The scope to initialise the objectlet as. Defaults to `singleton`.
        lazy:
            If true, the objectlet is not initialised until it is first autowired in by something else.
            If false, as per the default, then it will be initialised immediately.
    """

    def __init__(self, *, scope: scopes.Scope = scopes.Scope.SINGLETON, lazy: bool = False):
        super().__init__()
        self.scope = scope
        self.lazy = lazy



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Dependency injector implementation.
"""
__all__ = ['DependencyInjector']

from typing import Any, Iterator

from ricochet import autowired
from ricochet import dependency
from ricochet import objectlet_registries
from ricochet.utils import logger_mixin

from . import base_dependency_injector


class DependencyInjector(base_dependency_injector.BaseDependencyInjector, logger_mixin.LoggerMixin):
    """
    Basic implementation of a dependency injector,
    """

    def __init__(self, registry: objectlet_registries.BaseRegistry):
        super().__init__()
        self.registry = registry

    def resolve_dependencies(self, object: Any) -> None:
        dependencies = self.get_dependencies(object)
        for dependency_descriptor in dependencies:
            if isinstance(dependency_descriptor, dependency.ObjectletDependencyDescriptor):
                objectlet = self.get_objectlet_from_dependency(dependency_descriptor)
                self.wire_in_objectlet(object, dependency_descriptor.name, objectlet)
            else:
                raise NotImplementedError(
                    f'No handler for {type(dependency_descriptor).__name__} is implemented yet.'
                )

    def get_dependencies(self, object: Any) -> Iterator[dependency.BaseDependencyDescriptor]:
        all_annotations: dict = getattr(type(object), '__annotations__', {})

        for name, annotation in all_annotations.items():
            if isinstance(annotation, type) and issubclass(annotation, autowired.Autowired):
                self.logger.debug('Discovered requirement %r in %s', name, object)
                yield dependency.ObjectletDependencyDescriptor(name=name)

    def wire_in_objectlet(self, object: Any, name: str, objectlet: Any) -> None:
        self.logger.debug('Autowiring %r into %s', name, object)
        setattr(object, name, objectlet)

    def get_objectlet_from_dependency(
        self, dependency_descriptor: dependency.ObjectletDependencyDescriptor
    ) -> Any:
        canonical_name = self.registry.get_canonical_name_for(dependency_descriptor.name)
        return self.registry.get_objectlet(canonical_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implementation of a :class:`ricochet.objectlet_registries.prototype_and_singleton_registry.BaseDualRegistryMediator`
that uses the basic prototype registry and basic singleton registry internally.
"""
__all__ = ["DualRegistryMediator"]

from ricochet.objectlet_registries import base_dual_registry_mediator
from ricochet.objectlet_registries.impl import prototype_registry, singleton_registry


class DualRegistryMediator(
    base_dual_registry_mediator.BaseDualRegistryMediator,
    prototype_registry.PrototypeRegistry,
    singleton_registry.SingletonRegistry,
):
    """
    Implements the `BaseDualRegistryMediator` wrapper of a `PrototypeRegistry` and a `SingletonRegistry`.
    """

    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Represents a resolvable dependency.
"""
from typing import Any, Type

from ricochet.utils import hints


@hints.abstract_class
@hints.slotted
class BaseDependencyDescriptor:
    """Represents the base interface for a dependency descriptor."""

    __slots__ = ()

    def __init__(self):
        if type(self) == BaseDependencyDescriptor:
            raise TypeError()


@hints.slotted
class ObjectletDependencyDescriptor(BaseDependencyDescriptor):
    """
    Represents a dependency to resolve.
    """

    __slots__ = ('name',)

    def __init__(self, *, name: str) -> None:
        super().__init__()
        self.name = name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

PACKAGE = "ricochet"

# These are added by CI automatically.
__version__ = "0.2.1"


# Fill in manually.
__requirements__ = ["metabeyond>=0.2.0"]
__author__ = "flitt3r"
__license__ = "MIT"


# Used by CI, fill in manually...
__deploy_requirements__ = ["twine", "black", "wheel", "setuptools"]
__test_requirements__ = ["coverage", "asynctest"]
__sast_requirements__ = ["bandit"]
__lint_requirements__ = ["pylint"]
__docs_requirements__ = [
    "sphinx",
    "graphviz",
    "Jinja2",
    "solar-theme",
    "sphinxcontrib-fulltoc",
    "sphinx-autodoc-typehints",
    "sphinxcontrib-asyncio",
]


# Only works when building, not installing...
try:
    with open('README.md') as fp:
        print('Read in the README')
        readme_content = fp.read()
    readme_content_type = 'text/markdown'
except Exception:
    readme_content = 'Visit the GitLab repo for more details'
    readme_content_type = 'text/plain'


# Decent reference:
# https://gemfury.com/stevenferreira/python:sample/-/content/setup.py?gclid=Cj0KCQiApILhBRD1ARIsAOXWTztNWw_wmu2xA4Ka0yn-EjwbS99NZBI19Pjozfw7qAjVGSXDTY6grkkaAm1WEALw_wcB
setup(
    name=PACKAGE,
    version=__version__,
    description="Simple Spring-inspired IoC framework for Python.",
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Development Status :: 1 - Planning
        # Development Status :: 2 - Pre-Alpha
        # Development Status :: 3 - Alpha
        # Development Status :: 4 - Beta
        # Development Status :: 5 - Production/Stable
        # Development Status :: 6 - Mature
        # Development Status :: 7 - Inactive
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="python ioc spring inversion of control",
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    install_requires=__requirements__,
    extras_require={
        "test": __test_requirements__,
        "lint": __lint_requirements__,
        "sast": __sast_requirements__,
        "docs": __docs_requirements__,
        "deploy": __deploy_requirements__,
        "develop": __requirements__,
    },
    include_package_data=True,
    python_requires=">=3.7.0",
    long_description=readme_content,
    long_description_content_type=readme_content_type,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Defines a base registry interface for a prototype registry implementation.

**Prototype objects**:
    An object which is stored internally by the registry once. Each component that requests a reference to the
    prototype will receive some object created from the state of the prototype. Changing this object will not
    mutate the original prototype in any way, and thus provides a copy to every component that requires one.

    This is very similar to a normal instance, but instead of using a class to create an object, we use an
    existing object to derive a copy from.

    Likewise, these are much more difficult to track, as there is no implementation for tracking instances
    created from prototypes.
"""

__all__ = ["BasePrototypeRegistry"]

import abc
from typing import Any, KeysView, Optional, Type, Iterator


from ricochet import scopes
from ricochet import exceptions
from ricochet.utils import hints
from . import base_registry


@hints.abstract_class
class BasePrototypeRegistry(base_registry.BaseRegistry):
    """
    Defines an object which maintains a collection of prototype objectlets. Prototype objectlets are instances of a
    class that exist as a template for producing copies for use at runtime. This allows us to encapsulate the state of
    an existing object and pass copies to multiple components across the application context.

    .. image:: resources_static/prototype-cardinality.png

    If one component changes the state of a prototype copy, it will not affect other components using copies of the
    same prototype. Likewise, an instance made from a prototype may only belong to a single object. This owning object
    is allowed to have multiple instances of the same prototype objectlet if desired.

    .. image:: resources_static/prototype-distribution.png

    Note:
        Due to the issue that keeping references to objects prevents their destruction, and that it is dangerous to
        provide direct access to weak reference objects as it is non deterministic as to whether they get garbage
        collected halfway through being used by an introspective algorithm, no interface for accessing instances of
        prototypes is implemented here. However, whether this is implemented or not is entirely up to you if you derive
        your own concrete implementations of this class.

    Note:
        The method of copying objectlets is determined by the implementation.

    Takes a generic TypeVar :attr:`PrototypeWrapperT` describing the base type to output when queried for a prototype
    instance. If no wrapper is provided, this can be omitted, in which case, it will default to `Any`.

    Warning:
        This interface is designed to not expose the prototype bases themselves once registered. If this behaviour
        is desired, then it is down to the implementation to provide this, however this is generally not recommended
        unless a good use case can be made.
        """

    @abc.abstractmethod
    def contains_prototype_named(self, canonical_name: str) -> bool:
        """
        Args:
            canonical_name:
                The canonical objectlet name. This is the primary name the objectlet is registered under and resolved
                by. This implements no form of alias lookup.

        Returns:
            `True` if a prototype with the canonical name is registered, `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def contains_prototypes_typed(self, type: Type[Any]) -> bool:
        """
        Args:
            type:
                A class that a prototype is expected to be derived from.

        Returns:
            `True` if any prototype derived from the given type exists, `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def create_instance_of_prototype(self, canonical_name: str) -> Optional[Any]:
        """
        Args:
            canonical_name:
                The canonical prototype name. This is the primary name the objectlet is registered under and resolved
                by. This implements no form of alias lookup.

        Returns:
            TODO: change this to raise exception.
            A new instance of the prototype with the given name, or `None` if it doesn't exist...
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def register_new_prototype(self, canonical_name: str, prototype: Any) -> None:
        """
        Registers the given object as a prototype objectlet under the given name.

        Args:
           canonical_name:
               The canonical prototype's name. This is the primary name the objectlet is registered under and resolved
               by.
           prototype:
               The object itself to register.

        Throws:
           :class:`ValueError`:
               if the prototype is `None`.
           :class:`NameError`:
               if the prototype is already registered.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def is_prototype_instance(self, candidate_prototype_instance: Any) -> bool:
        """
        Returns:
             `True` if the given object is an instance of a registered prototype, `False` otherwise.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def count_instances_of_prototype_named(self, canonical_name: str) -> int:
        """
        Args:
            canonical_name:
                The canonical objectlet name. This is the primary name the object is registered under and resolved by.
                This implements no form of alias lookup.

        Returns:
            A count of prototype instances, if the prototype exists for that name.

        Raises:
            NameError if no prototype exists with that name.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_all_prototype_names(self) -> KeysView[str]:
        """
        Returns:
            View across all prototype names.
        """
        raise exceptions.AbstractError

    @property
    @abc.abstractmethod
    def prototype_count(self) -> int:
        """
        The number of objectlets registered as prototypes.
        """
        raise exceptions.AbstractError

    def get_objectlet(self, canonical_name: str) -> Any:
        return self.create_instance_of_prototype(canonical_name)

    def contains_objectlet_named(self, canonical_name: str) -> bool:
        return self.contains_prototype_named(canonical_name)

    def contains_objectlets_typed(self, type: Type[Any]) -> bool:
        return self.contains_prototypes_typed(type)

    def count_objectlets_named(self, canonical_name: str) -> int:
        return self.count_instances_of_prototype_named(canonical_name)

    def is_objectlet(self, candidate_object: Any) -> bool:
        return self.is_prototype_instance(candidate_object)

    def get_all_canonical_objectlet_names(self) -> Iterator[str]:
        yield from self.get_all_prototype_names()

    def register_new_objectlet(self, canonical_name: str, object: Any, scope: scopes.Scope) -> None:
        if scope is not scopes.Scope.PROTOTYPE:
            raise NotImplementedError('This class can only register prototype objects')
        else:
            return self.register_new_prototype(canonical_name, object)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Mocking utility for quickly stubbing an ABC that is being tested. This is useful for testing mixin classes that
are ABCs where you don't wish to automatically implement each one with a stub method, because that is a lot of
boilerplate code.
"""
import inspect
import textwrap

import asynctest


def _stub_method(method):
    name = method.__name__
    modifier = 'async def' if inspect.iscoroutinefunction(method) else 'def'
    signature = inspect.signature(method)
    params = signature.parameters

    safe_signature = '(' + ', '.join(params.keys()) + ')'

    mock_class = asynctest.MagicMock

    return (
        f'{modifier} {name}{safe_signature}:\n'
        '    """Overrides abstract method"""\n'
        f'    return {mock_class.__module__}.{mock_class.__name__}()\n'
    )


def _stub_property(name, property):
    modifier = 'async def' if inspect.iscoroutinefunction(property.getter) else 'def'
    safe_signature = '(self)'

    mock_class = asynctest.MagicMock

    return (
        '@property\n'
        f'{modifier} {name}{safe_signature}:\n'
        '    """Overrides abstract property getter"""\n'
        f'    return {mock_class.__module__}.{mock_class.__name__}()\n'
    )


def make_abc_mock_class(class_):
    """

    Args:
        class_: the class to stub ABC methods for.
        *args: arguments for the constructor of said class.
        **kwargs: keyword arguments for the constructor of said class.

    Returns:
        A stubbed class for quick mocking. Each method is a stub that returns some mock object immediately.

    Note:
        Asynchronous mocks are implemented using :mod:`asynctest`' coroutine mocks.
    """
    method_stubs = []

    for name, member in inspect.getmembers(class_):
        if not getattr(member, '__isabstractmethod__', False):
            continue
        if inspect.isdatadescriptor(member):
            stub = _stub_property(name, member)
        else:
            stub = _stub_method(member)

        method_stubs.append(stub)

    mock_name = f'{class_.__name__}Mock'

    mock_class_body = f'class {mock_name}(class_):\n' + '\n'.join(
        textwrap.indent(method, '    ') for method in method_stubs
    )

    exec(mock_class_body)  # nosec

    return locals()[mock_name]


def make_abc_mock_instance(class_, *args, **kwargs):
    """
    Args:
        class_: the class to stub ABC methods for.
        *args: arguments for the constructor of said class.
        **kwargs: keyword arguments for the constructor of said class.

    Returns:
        An instance of a stubbed class implementing mocks for each abstract method of the given class for quick mocking.

    Note:
        This will not currently correctly stub properties. Asynchronous mocks are implemented using :mod:`asynctest`'s
        coroutine mocks.
    """
    return make_abc_mock_class(class_)(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Used internally by other framework components to provide logging in a simple and consistent manner.
"""
import logging


__all__ = ["LoggerMixin"]


class LoggerMixin:
    """
    Injects a logger into class scope with an appropriate name.

    Objects will use a logger adapter facade to enable MDC-like behaviour per object.
    """

    logger: logging.Logger

    def __init_subclass__(cls, **_):
        cls.logger: logging.Logger = logging.getLogger(cls.__name__)

    def __init__(self):
        logger = logging.getLogger(type(self).__name__)
        extra = {"object_id": id(self)}
        self.logger: logging.LoggerAdapter = logging.LoggerAdapter(logger, extra=extra)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Represents a generic registry. This is an ABC that an application context can be expected to interface with.

There is no requirement to implement all of the methods, only some need to be implemented. This is designed to allow
a Singleton registry and a Prototype registry to interop together, or for only one implementation to be used.
"""

__all__ = ["BaseRegistry"]

import abc
from typing import Iterator, Type, Any, Optional

from ricochet import exceptions, scopes
from ricochet.utils import hints, logger_mixin


@hints.abstract_class
class BaseRegistry(abc.ABC, logger_mixin.LoggerMixin):
    """
    A generic type of registry that is expected to be the interface used in an application. A registry is some
    form of collection of objects that are managed by this framework, and are thus known as objectlets. This acts as
    a form of "objectlet repository" for dependency injection components that require access to a collection of
    objectlets at runtime under a restricted and managed interface.

    The importance of this to be managed in a correct way is increased with the dynamic and non-encapsulated nature
    of Python code.
    """

    def __init__(self):
        super().__init__()
        #: Maps aliases to names.
        self._aliases = {}

    @abc.abstractmethod
    def get_objectlet(self, canonical_name: str) -> Any:
        """
        Gets the given objectlet with the given canonical name, and type.

        Args:
            canonical_name:
                The objectlet's canonical name to find.

        Returns:
            The located objectlet.

        Raises:
            NameError if the objectlet cannot be found.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def register_new_objectlet(self, canonical_name: str, object: Any, scope: scopes.Scope) -> None:
        """
        Register a new objectlet as a given scope under the given canonical name.

        Args:
            canonical_name:
                The objectlet's canonical name to be registered under.
            object:
                The objectlet to register.
            scope:
                The scope to use for the object when resolving it.

        Raises:
            :cls:`exceptions.ObjectletIsAlreadyDefined` if an objectlet already exists with this name.
            :cls:`ValueError` if the given object is `None`.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def contains_objectlet_named(self, canonical_name: str) -> bool:
        """
        Args:
            canonical_name:
                The canonical objectlet name. This is the primary name the objectlet is registered under and resolved
                by. This implements no form of alias lookup.

        Returns:
            `True` if a prototype or singleton with the canonical name is registered, `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def contains_objectlets_typed(self, type: Type[Any]) -> bool:
        """
        Args:
            type:
                A class that a prototype or singleton is expected to be derived from.

        Returns:
            `True` if any prototype or singleton derived from the given type exists, `False` if not.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def count_objectlets_named(self, canonical_name: str) -> int:
        """
        Args:
            canonical_name:
                The canonical object name. This is the primary name the object is registered under and resolved by.
                This implements no form of alias lookup.

        Returns:
            A count of object instances, if the prototype or singleton exists for that name.

        Raises:
            NameError if no prototype or singleton exists with that name.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def is_objectlet(self, candidate_object: Any) -> bool:
        """
        Args:
            candidate_object: The potential singleton or prototype instance to look for.

        Returns:
            `True` if this object is a prototype instance or a singleton objectlet in this registry, or `False`
            otherwise.
        """
        raise exceptions.AbstractError

    @abc.abstractmethod
    def get_all_canonical_objectlet_names(self) -> Iterator[str]:
        """
        Yields all objectlet names. These are expected to be unique as long as the internal representations that are
        protected have not been tampered with directly.
        """
        raise exceptions.AbstractError

    def get_all_objectlet_names(self) -> Iterator[str]:
        """
        Returns:
            The same as :attr:`get_all_canonical_objectlet_names`, but also any stored aliases. The order of these
            is not defined, and should not be relied upon.
        """
        yield from self.get_all_canonical_objectlet_names()
        yield from self._aliases

    def get_canonical_name_for(self, alias: str) -> Optional[Any]:
        """
        Looks up an alias or canonical name and returns the matching canonical name for an objectlet.

        Args:
            alias:
                any alias for an objectlet.

        Returns:
            The alias input if it is already a canonical name (primary name for an objectlet). If it is a valid
            alias, the canonical name will always be returned. If the alias is not registered, then we instead return
            None.
        """
        if alias in self.get_all_canonical_objectlet_names():
            return alias

        try:
            return self._aliases[alias]
        except KeyError:
            return None

    def get_aliases_for(self, alias: str) -> Iterator[str]:
        """
        Gets the aliases for a given canonical name or alias as an iterator.

        Args:
            alias:
                The canonical name or alias to look up

        Returns:
            An iterator across all aliases. This does not include the canonical name. If no alias with the provided
            name exists, then no result will be yielded.

            To determine if an alias exists, one should use the :meth:`get_canonical_name_for` method instead.
        """
        canonical_name = self.get_canonical_name_for(alias)
        if canonical_name is not None:
            for alias, name in self._aliases.items():
                if name == canonical_name:
                    yield alias

    def get_names_for(self, alias: str) -> Iterator[str]:
        """
        Args:
            alias:
                Alias or canonical name to look up.

        Returns:
            An iterator across all canonical and object names for the given alias/canonical name. If no match
            is found, then no items are yielded. This order is arbitrary.
        """
        canonical_name = self.get_canonical_name_for(alias)
        if canonical_name is not None:
            yield canonical_name
            yield from self.get_aliases_for(canonical_name)

    def register_alias(self, new_alias, existing_alias) -> None:
        """
        Registers an alias to an existing object alias, thus allowing it to be referred to by the new alias
        additionally.

        Args:
            new_alias:
                The new alias to register and use to refer to an object.
            existing_alias:
                The name of the existing object to apply the new alias to. This may be an alias or a canonical name.

        Throws:
            ValueError if the existing alias does not exist.
            NameError if the new alias already exists.
        """

        if new_alias in self.get_all_objectlet_names():
            canonical_name = self.get_canonical_name_for(new_alias)
            # Assume this always matches, this just helps provide a decent error message.
            raise exceptions.ObjectletAliasIsAlreadyTakenError(
                new_alias, existing_alias, canonical_name
            )
        else:
            canonical_name = self.get_canonical_name_for(existing_alias)
            if canonical_name is None:
                raise exceptions.ObjectletNotFoundError(existing_alias)
            else:
                self._aliases[new_alias] = canonical_name

    def contains_objectlet_with_alias(self, candidate_name: str) -> bool:
        """
        Return true if the name is an alias, but not a canonical name.
        """
        return candidate_name in self.get_aliases_for(candidate_name)

    def contains_objectlet_with_canonical_name(self, candidate_name: str) -> bool:
        """
        Return true if the name is a valid canonical name.
        """
        return candidate_name in self.get_all_canonical_objectlet_names()

    @property
    def alias_count(self) -> int:
        """
        Number of registered aliases not including canonical names.
        """
        return len(self._aliases)

    @property
    def name_count(self) -> int:
        """
        Number of registered aliases and canonical names.
        """
        return sum(1 for _ in self.get_all_objectlet_names())

    @property
    def objectlet_count(self) -> int:
        """
        Returns:
            The number of objectlets that are managed.
        """
        return sum(1 for _ in self.get_all_canonical_objectlet_names())
